"""
Signal generation stack.

Contrary to most quanutm simulators, C^3 includes a detailed simulation of the control
stack. Each component in the stack and its functions are simulated individually and
combined here.

Example: A local oscillator and arbitrary waveform generator signal
are put through via a mixer device to produce an effective modulated signal.
"""

from typing import List, Callable, Dict
import hjson
import tensorflow as tf
from c3.c3objs import hjson_decode, hjson_encode
from c3.signal.gates import Instruction
from c3.generator.devices import devices as dev_lib
from c3.generator.devices import Device


class Generator:
    """
    Generator, creates signal from digital to what arrives to the chip.

    Parameters
    ----------
    devices : list
        Physical or abstract devices in the signal processing chain.
    resolution : np.float64
        Resolution at which continuous functions are sampled.
    callback : Callable
        Function that is called after each device in the signal line.

    """

    def __init__(
        self,
        devices: dict = None,
        chains: dict = None,
        resolution: float = 0.0,
        callback: Callable = None,
    ):
        self.devices: Dict[str, Device] = {}
        if devices:
            self.devices = devices
        self.chains: Dict[str, Dict[str, List[str]]] = {}
        self.sorted_chains: Dict[str, List[str]] = {}
        self.set_chains(chains)
        self.resolution = resolution
        self.callback = callback

    def set_chains(self, chains: Dict[str, Dict[str, List[str]]]):
        if chains:
            self.chains = chains
            self.__check_signal_chains()

    def __check_signal_chains(self) -> None:
        for channel, chain in self.chains.items():
            signals = 0
            for device_id, sources in chain.items():
                # all source devices need to exist and have the same resolution
                if sources:
                    res = self.devices[sources[0]].resolution
                for dev in sources:
                    if dev not in self.devices:
                        raise Exception(f"C3:Error: device {dev} not found.")
                    if res != self.devices[dev].resolution:
                        raise Exception(
                            f"C3:Error: Different resolution of inputs in {channel} {device_id}:{sources}."
                        )

                # the expected number of inputs must match the connected devices
                if self.devices[device_id].inputs != len(sources):
                    raise Exception(
                        f"C3:Error: device {device_id} expects {self.devices[device_id].inputs} inputs, but {len(sources)} found."
                    )

                # overall the chain should have exactly 1 output signal
                signals -= self.devices[device_id].inputs
                signals += self.devices[device_id].outputs
            if signals != 1:
                raise Exception(
                    "C3:ERROR: Signal chain for channel '"
                    + channel
                    + "' contains unmatched number of inputs and outputs."
                )

            # bring chain in topological order
            self.sorted_chains[channel] = self.__topological_ordering(
                self.chains[channel]
            )

    def __topological_ordering(self, predecessors: Dict[str, List[str]]) -> List[str]:
        """
        Computes the topological ordering of a directed acyclic graph.

        Parameters
        ----------
        predecessors : dict
            list of preceding nodes for each node

        Returns
        -------
            a list of all nodes in topological ordering

        Raises
        ------
        ValueError
            if the graph contains a cycle
        """
        stack = [x for x in predecessors if len(predecessors[x]) == 0]
        num_sources = {node: len(predecessors[node]) for node in predecessors}
        successors = {}
        for node in predecessors:
            successors[node] = [x for x in predecessors if node in predecessors[x]]
        ordered = []

        while stack:
            src = stack.pop()
            for node in successors[src]:
                num_sources[node] -= 1
                if num_sources[node] == 0:
                    stack.append(node)
            ordered.append(src)

        if len(ordered) != len(successors):
            raise Exception("C3:ERROR: Device chain contains a cycle")
        return ordered

    def read_config(self, filepath: str) -> None:
        """
        Load a file and parse it to create a Generator object.

        Parameters
        ----------
        filepath : str
            Location of the configuration file

        """
        with open(filepath, "r") as cfg_file:
            cfg = hjson.loads(cfg_file.read(), object_pairs_hook=hjson_decode)
        self.fromdict(cfg)

    def fromdict(self, cfg: dict) -> None:
        for name, props in cfg["Devices"].items():
            dev_type = props.pop("c3type")
            if "name" not in props:
                props["name"] = name
            self.devices[name] = dev_lib[dev_type](**props)
        self.chains = cfg["Chains"]
        self.__check_signal_chains()

    def write_config(self, filepath: str) -> None:
        """
        Write dictionary to a HJSON file.
        """
        with open(filepath, "w") as cfg_file:
            hjson.dump(self.asdict(), cfg_file, default=hjson_encode)

    def asdict(self) -> dict:
        """
        Return a dictionary compatible with config files.
        """
        devices = {}
        for name, dev in self.devices.items():
            devices[name] = dev.asdict()
        return {"Devices": devices, "Chains": self.chains}

    def __str__(self) -> str:
        return hjson.dumps(self.asdict(), default=hjson_encode)

    def generate_signals(self, instr: Instruction) -> dict:
        """
        Perform the signal chain for a specified instruction, including local
        oscillator, AWG generation and IQ mixing.

        Parameters
        ----------
        instr : Instruction
            Operation to be performed, e.g. logical gate.

        Returns
        -------
        dict
            Signal to be applied to the physical device.

        """
        gen_signal: Dict[str, Dict[str, tf.constant]] = {}
        signal_stack: Dict[str, Dict[str, tf.constant]] = {}
        for chan in instr.comps:
            chain = self.chains[chan]
            signal_stack[chan] = {}

            # create list of succeeding devices
            successors = {}
            for dev_id in chain:
                successors[dev_id] = [x for x in chain if dev_id in chain[x]]

            for dev_id in self.sorted_chains[chan]:
                # collect inputs
                sources = self.chains[chan][dev_id]
                inputs = [signal_stack[chan][x] for x in sources]

                # calculate the output and store it in the stack
                dev = self.devices[dev_id]
                output = dev.process(instr, chan, inputs)
                signal_stack[chan][dev_id] = output

                # remove inputs if they are not needed anymore
                # for source in sources:
                #     successors[source].remove(dev_id)
                #     if len(successors[source]) < 1:
                #         del signal_stack[chan][source]

                # call the callback with the current signal
                if self.callback:
                    self.callback(chan, dev_id, output)

            gen_signal[chan] = {}
            for key in output.keys():
                gen_signal[chan][key] = tf.identity(output[key])

        # Hack to use crosstalk. Will be generalized to a post-processing module.
        # TODO: Rework of the signal generation for larger chips, similar to qiskit
        if "crosstalk" in self.devices:
            gen_signal = self.devices["crosstalk"].process(
                None, None, signals=[gen_signal]
            )
        return gen_signal



"""ParameterMap class"""

from typing import List, Dict, Tuple
import hjson
import copy
import numpy as np
import tensorflow as tf
from c3.c3objs import Quantity, hjson_decode, hjson_encode
from c3.signal.gates import Instruction
from typing import Union
from tensorflow.errors import InvalidArgumentError


class ParameterMap:
    """
    Collects information about control and model parameters and provides different
    representations depending on use.
    """

    def __init__(
        self, instructions: List[Instruction] = [], generator=None, model=None
    ):
        self.instructions: Dict[str, Instruction] = dict()
        self.opt_map: List[List[Tuple[str]]] = list()
        self.model = model
        self.generator = generator
        for instr in instructions:
            self.instructions[instr.get_key()] = instr

        # Collecting model components
        components = {}
        if model:
            components.update(model.couplings)
            components.update(model.subsystems)
            components.update(model.tasks)
        if generator:
            components.update(generator.devices)
        self.__components = components
        self.update_model = False
        self.set_parameters_scaled = self._set_parameters_scaled_ctrls
        self.__initialize_parameters()

    def __initialize_parameters(self) -> None:
        par_lens = {}
        pars = {}
        par_ids_model = []
        for comp in self.__components.values():
            for par_name, par_value in comp.params.items():
                par_id = "-".join([comp.name, par_name])
                par_lens[par_id] = par_value.length
                pars[par_id] = par_value
                par_ids_model.append(par_id)

        # Initializing control parameters
        for gate in self.instructions:
            instr = self.instructions[gate]
            for key_elems, par_value in instr.get_optimizable_parameters():
                par_id = "-".join(key_elems)
                par_lens[par_id] = par_value.length
                pars[par_id] = par_value

        self._par_lens = par_lens
        self._pars: Dict[str, Quantity] = pars
        self._par_ids_model = par_ids_model

    def update_parameters(self):
        self.__initialize_parameters()

    def load_values(self, init_point):
        """
        Load a previous parameter point to start the optimization from.

        Parameters
        ----------
        init_point : str
            File location of the initial point

        """
        with open(init_point) as init_file:
            best = hjson.load(init_file, object_pairs_hook=hjson_decode)

        best_opt_map = best["opt_map"]
        init_p = best["optim_status"]["params"]
        self.set_parameters(init_p, best_opt_map, extend_bounds=True)

    def store_values(self, path: str, optim_status=None) -> None:
        """
        Write current parameter values to file. Stores the numeric values, as well as the names
        in form of the opt_map and physical units. If an optim_status is given that will be
        used.

        Parameters
        ----------
        path : str
            Location of the resulting logfile.
        optim_status: dict
            Dictionary containing current parameters and goal function value.
        """
        if optim_status is None:
            optim_status = {
                "params": [par.numpy().tolist() for par in self.get_parameters()]
            }
        with open(path, "w") as value_file:
            val_dict = {
                "opt_map": self.get_opt_map(),
                "units": self.get_opt_units(),
                "optim_status": optim_status,
            }
            value_file.write(hjson.dumps(val_dict, default=hjson_encode))
            value_file.write("\n")

    def read_config(self, filepath: str) -> None:
        """
        Load a file and parse it to create a ParameterMap object.

        Parameters
        ----------
        filepath : str
            Location of the configuration file

        """
        with open(filepath, "r") as cfg_file:
            cfg = hjson.loads(cfg_file.read(), object_pairs_hook=hjson_decode)
        self.fromdict(cfg)

    def fromdict(self, cfg: dict) -> None:
        for key, gate in cfg.items():
            if "mapto" in gate.keys():
                instr = copy.deepcopy(self.instructions[gate["mapto"]])
                instr.name = key
                for drive_chan, comps in gate["drive_channels"].items():
                    for comp, props in comps.items():
                        for par, val in props["params"].items():
                            instr.comps[drive_chan][comp].params[par].set_value(val)
            else:
                # TODO: initialize directly by using the constructor.
                instr = Instruction()
                instr.from_dict(gate, name=key)
            self.instructions[key] = instr
            self.__initialize_parameters()

    def write_config(self, filepath: str) -> None:
        """
        Write dictionary to a HJSON file.
        """
        with open(filepath, "w") as cfg_file:
            hjson.dump(self.asdict(), cfg_file, default=hjson_encode)

    def asdict(self, instructions_only=True) -> dict:
        """
        Return a dictionary compatible with config files.
        """
        instructions = {}
        for name, instr in self.instructions.items():
            instructions[name] = instr.asdict()
        if instructions_only:
            return instructions
        else:
            out_dict = dict()
            out_dict["instructions"] = instructions
            out_dict["model"] = self.model.asdict()
            return out_dict

    def __str__(self) -> str:
        return hjson.dumps(self.asdict(), default=hjson_encode)

    def get_full_params(self) -> Dict[str, Quantity]:
        """
        Returns the full parameter vector, including model and control parameters.
        """
        return self._pars

    def get_not_opt_params(self, opt_map=None) -> Dict[str, Quantity]:
        opt_map = self.get_opt_map(opt_map)
        out_dict = copy.copy(self._pars)
        for equiv_ids in opt_map:
            for key in equiv_ids:
                del out_dict[key]
        return out_dict

    def get_opt_units(self) -> List[str]:
        """
        Returns a list of the units of the optimized quantities.
        """
        units = []
        for equiv_ids in self.get_opt_map():
            key = equiv_ids[0]
            units.append(self._pars[key].unit)
        return units

    def get_opt_limits(self):
        limits = []
        for equiv_ids in self.get_opt_map():
            key = equiv_ids[0]
            limits.append((self._pars[key].get_limits()))
        return limits

    def check_limits(self, opt_map):
        """
        Check if all elements of equal ids have the same limits. This has to be checked against if setting values optimizer friendly.

        Parameters
        ----------
        opt_map

        Returns
        -------

        """
        for equiv_ids in self.get_opt_map():
            if len(equiv_ids) > 1:
                limit = self._pars[equiv_ids[0]].get_limits()
                for key in equiv_ids[1:]:
                    if not self._pars[key].get_limits() == limit:
                        raise Exception(
                            "C3:Error:Limits for {key} are not equivalent to {equiv_ids}."
                        )

    def get_parameter(self, par_id: Tuple[str, ...]) -> Quantity:
        """
        Return one the current parameters.

        Parameters
        ----------
        par_id: tuple
            Hierarchical identifier for parameter.

        Returns
        -------
        Quantity

        """
        key = "-".join(par_id)
        try:
            value = self._pars[key]
        except KeyError as ke:
            raise Exception(f"C3:ERROR:Parameter {key} not defined.") from ke
        return value

    def get_parameters(self, opt_map=None) -> List[Quantity]:
        """
        Return the current parameters.

        Parameters
        ----------
        opt_map: list
            Hierarchical identifier for parameters.

        Returns
        -------
        list of Quantity

        """
        values = []
        opt_map = self.get_opt_map(opt_map)
        for equiv_ids in opt_map:
            key = equiv_ids[0]
            values.append(self._pars[key])
        return values

    def get_parameter_dict(self, opt_map=None) -> Dict[str, Quantity]:
        """
        Return the current parameters in a dictionary including keys.
        Parameters
        ----------
        opt_map

        Returns
        -------
        Dictionary with Quantities
        """
        value_dict = dict()
        opt_map = self.get_opt_map(opt_map)
        for equiv_ids in opt_map:
            key = equiv_ids[0]
            value_dict[key] = self._pars[key]
        return value_dict

    def set_parameters(
        self, values: Union[List, np.ndarray], opt_map=None, extend_bounds=False
    ) -> None:
        """Set the values in the original instruction class.

        Parameters
        ----------
        values: list
            List of parameter values. Can be nested, if a parameter is matrix valued.
        opt_map: list
            Corresponding identifiers for the parameter values.
        extend_bounds: bool
            If true bounds of quantity objects will be extended.

        """
        model_updated = False
        val_indx = 0
        opt_map = self.get_opt_map(opt_map)
        if not len(values) == len(opt_map):
            raise Exception(
                f"C3:Error: Different number of elements in values and opt_map. {len(values)} vs {len(opt_map)}"
            )
        for equiv_ids in opt_map:
            for key in equiv_ids:
                # We check if a model parameter has changed
                model_updated = key in self._par_ids_model or model_updated
                try:
                    par = self._pars[key]
                except ValueError as ve:
                    raise Exception(f"C3:ERROR:{key} not defined.") from ve
                try:
                    par.set_value(values[val_indx], extend_bounds=extend_bounds)
                except (ValueError, InvalidArgumentError) as ve:
                    try:
                        raise Exception(
                            f"C3:ERROR:Trying to set {key} "
                            f"to value {values[val_indx]} "
                            f"but has to be within {par.offset} .."
                            f" {(par.offset + par.scale)}."
                        ) from ve
                    except TypeError:
                        raise ve
            val_indx += 1

        # TODO: This check is too simple. Not every model parameter requires an update.
        if model_updated and self.model:
            self.model.update_model()

    def get_parameters_scaled(self, opt_map=None) -> np.ndarray:
        """
        Return the current parameters. This fuction should only be called by an
        optimizer. Are you an optimizer?

        Parameters
        ----------
        opt_map: tuple
            Hierarchical identifier for parameters.

        Returns
        -------
        list of Quantity

        """
        values = []
        opt_map = self.get_opt_map(opt_map)
        for equiv_ids in opt_map:
            key = equiv_ids[0]
            par = self._pars[key]
            values.append(par.get_opt_value())
        return values

    def _set_parameters_scaled_ctrls(
        self, values: Union[tf.constant, tf.Variable], opt_map=None
    ) -> None:
        """
        Set the values in the original instruction class. This fuction should only be
        called by an optimizer. Are you an optimizer? This method only sets control
        parameters and does not trigger a model update.

        Parameters
        ----------
        values: list
            List of parameter values. Matrix valued parameters need to be flattened.

        """
        val_indx = 0
        opt_map = self.get_opt_map(opt_map)
        for equiv_ids in opt_map:
            key = equiv_ids[0]
            par_len = self._pars[key].length
            for par_id in equiv_ids:
                key = par_id
                par = self._pars[key]
                par.set_opt_value(values[val_indx : val_indx + par_len])
            val_indx += par_len

    def _set_parameters_scaled_model(
        self, values: Union[tf.constant, tf.Variable], opt_map=None
    ) -> None:
        """
        Set the values in the original instruction class. This fuction should only be
        called by an optimizer. Are you an optimizer? Also update the model.

        Parameters
        ----------
        values: list
            List of parameter values. Matrix valued parameters need to be flattened.

        """
        self._set_parameters_scaled_ctrls(values)
        self.model.update_model()

    def get_key_from_scaled_index(self, idx, opt_map=None) -> str:
        """
        Get the key of the value at position `ìdx` of the scaled_parameters output
        Parameters
        ----------
        idx
        opt_map

        Returns
        -------

        """
        opt_map = self.get_opt_map(opt_map)
        curr_indx = 0
        for equiv_ids in opt_map:
            key = equiv_ids[0]
            par_len = self._pars[key].length
            curr_indx += par_len
            if idx < curr_indx:
                return key
        return ""

    def set_opt_map(self, opt_map) -> None:
        """
        Set the opt_map, i.e. which parameters will be optimized.
        """
        opt_map = self.get_opt_map(opt_map)
        update_model = False
        for equiv_ids in opt_map:
            for pid in equiv_ids:
                key = pid
                if key not in self._pars:
                    par_strings = "\n".join(self._pars.keys())
                    raise Exception(
                        f"C3:ERROR:Parameter {key} not defined in {par_strings}"
                    )
                update_model = key in self._par_ids_model or update_model
        self.set_update_model(update_model)
        self.check_limits(opt_map)
        self.opt_map = opt_map

    def set_update_model(self, update: bool) -> None:
        self.update_model = update
        if update:
            self.set_parameters_scaled = self._set_parameters_scaled_model
        else:
            self.set_parameters_scaled = self._set_parameters_scaled_ctrls

    def get_opt_map(self, opt_map=None) -> List[List[str]]:
        if opt_map is None:
            opt_map = self.opt_map

        for i, equiv_ids in enumerate(opt_map):
            for j, par_id in enumerate(equiv_ids):
                if type(par_id) is str:
                    continue
                key = "-".join(par_id)
                opt_map[i][j] = key

        return opt_map

    def str_parameters(
        self, opt_map: Union[List[List[Tuple[str]]], List[List[str]]] = None
    ) -> str:
        """
        Return a multi-line human-readable string of the optmization parameter names and
        current values.

        Parameters
        ----------
        opt_map: list
            Optionally use only the specified parameters.

        Returns
        -------
        str
            Parameters and their values
        """
        opt_map = self.get_opt_map(opt_map)
        ret = []
        for equiv_ids in opt_map:
            par_id = equiv_ids[0]
            key = par_id
            par = self._pars[key]
            ret.append(f"{key:38}: {par}\n")
            if len(equiv_ids) > 1:
                for eid in equiv_ids[1:]:
                    ret.append(eid)
                    ret.append("\n")
                ret.append("\n")
        return "".join(ret)

    def print_parameters(self, opt_map=None) -> None:
        """
        Print current parameters to stdout.
        """
        opt_map = self.get_opt_map(opt_map)
        print(self.str_parameters(opt_map))



"""Optimizer object, where the optimal control is done."""

import os
import time
from typing import Callable, Union, List, Dict, Optional, Any
import numpy as np
import tensorflow as tf
import hjson
import c3.libraries.algorithms as algorithms
from c3.c3objs import hjson_encode
from c3.experiment import Experiment
from c3.parametermap import ParameterMap
import copy
from tensorboard.plugins.hparams import api as hp
import warnings


class Optimizer:
    """
    General optimizer class from which specific classes are inherited.

    Parameters
    ----------
    algorithm : callable
        From the algorithm library
    store_unitaries : boolean
        Store propagators as text and pickle
    logger: List
        Logging classes
    """

    def __init__(
        self,
        pmap: ParameterMap,
        initial_point: str = "",
        algorithm: Callable = None,
        store_unitaries: bool = False,
        logger: List = None,
    ):
        self.pmap = pmap
        self.optim_status: Dict[str, Any] = dict()
        self.gradients: Dict[str, np.ndarray] = {}
        self.current_best_goal = 9876543210.123456789
        self.current_best_params = None
        self.evaluation = 0
        self.store_unitaries = store_unitaries
        self.created_by = None
        self.logname: str = ""
        self.options = None
        self.__dir_path: str = ""
        self.logdir: str = ""
        self.set_algorithm(algorithm)
        if not initial_point == "":
            self.load_best(initial_point)
        self.logger = []
        if logger is not None:
            self.logger = logger

    def set_algorithm(self, algorithm: Optional[Callable]) -> None:
        if algorithm:
            self.algorithm = algorithm
        else:
            print("C3:WARNING:No algorithm passed. Using default LBFGS")
            self.algorithm = algorithms.lbfgs

    def replace_logdir(self, new_logdir):
        """
        Specify a new filepath to store the log.

        Parameters
        ----------
        new_logdir

        """
        new_logdir = new_logdir.replace(":", "_")
        old_logdir = self.logdir
        self.logdir = new_logdir

        if old_logdir is None:
            return

        try:
            os.remove(os.path.join(self.__dir_path, "recent"))
        except FileNotFoundError:
            pass

        try:
            os.rmdir(old_logdir)
        except OSError:
            pass

        for logger in self.logger:
            logger.set_logdir

    def set_exp(self, exp: Experiment) -> None:
        self.exp = exp

    def set_created_by(self, config) -> None:
        """
        Store the config file location used to created this optimizer.
        """
        self.created_by = config

    def load_best(self, init_point) -> None:
        """
        Load a previous parameter point to start the optimization from. Legacy wrapper.
        Method moved to Parametermap.

        Parameters
        ----------
        init_point : str
            File location of the initial point

        """
        self.pmap.load_values(init_point)

    def start_log(self) -> None:
        """
        Initialize the log with current time.

        """
        self.start_time = time.time()
        start_time_str = str(f"{time.asctime(time.localtime())}\n\n")
        with open(self.logdir + self.logname, "a") as logfile:
            logfile.write("Starting optimization at ")
            logfile.write(start_time_str)
            logfile.write("Optimization parameters:\n")
            logfile.write(hjson.dumpsJSON(self.pmap.opt_map, default=hjson_encode))
            logfile.write("\n")
            logfile.write("Units:\n")
            logfile.write(
                hjson.dumpsJSON(self.pmap.get_opt_units(), default=hjson_encode)
            )
            logfile.write("\n")
            logfile.write("Algorithm options:\n")
            logfile.write(hjson.dumpsJSON(self.options, default=hjson_encode))
            logfile.write("\n")
            logfile.flush()

        for logger in self.logger:
            logger.start_log(self, self.logdir)

    def end_log(self) -> None:
        """
        Finish the log by recording current time and total runtime.

        """
        self.end_time = time.time()
        with open(self.logdir + self.logname, "a") as logfile:
            logfile.write(f"Finished at {time.asctime(time.localtime())}\n")
            logfile.write(f"Total runtime: {self.end_time - self.start_time}\n\n")
            logfile.flush()

        for logger in self.logger:
            logger.end_log(self, self.logdir)

    def log_best_unitary(self) -> None:
        """
        Save the best unitary in the log.
        """
        with open(self.logdir + "best_point_" + self.logname, "w") as best_point:
            propagators = self.exp.propagators
            for gate, U in propagators.items():
                best_point.write("\n")
                best_point.write(f"Re {gate}: \n")
                best_point.write(f"{np.round(np.real(U), 3)}\n")
                best_point.write("\n")
                best_point.write(f"Im {gate}: \n")
                best_point.write(f"{np.round(np.imag(U), 3)}\n")

    def log_parameters(self, params) -> None:
        """
        Log the current status. Write parameters to log. Update the current best
        parameters. Call plotting functions as set up.

        """
        if self.optim_status["goal"] < self.current_best_goal:
            self.current_best_goal = self.optim_status["goal"]
            self.current_best_params = self.optim_status["params"]
            self.pmap.store_values(
                path=self.logdir + "best_point_" + self.logname,
                optim_status=self.optim_status,
            )
        if self.store_unitaries:
            self.exp.store_Udict(self.optim_status["goal"])
            self.exp.store_unitaries_counter += 1

        with open(self.logdir + self.logname, "a") as logfile:
            logfile.write(
                f"\nFinished evaluation {self.evaluation} at {time.asctime()}\n"
            )
            # logfile.write(hjson.dumpsJSON(self.optim_status, indent=2))
            logfile.write(hjson.dumpsJSON(self.optim_status, default=hjson_encode))
            logfile.write("\n")
            logfile.flush()

        for logger in self.logger:
            logger.log_parameters(self.evaluation, self.optim_status)

    def goal_run(
        self, current_params: Union[np.ndarray, tf.constant]
    ) -> Union[np.ndarray, tf.constant]:
        """
        Placeholder for the goal function. To be implemented by inherited classes.
        """
        raise NotImplementedError("Implement this function in a subclass")

    def goal_run_with_grad(self, current_params):
        with tf.GradientTape(watch_accessed_variables=False) as t:
            t.watch(current_params)
            goal = self.goal_run(current_params)
        grad = t.gradient(goal, current_params)
        return goal, grad

    def lookup_gradient(self, x):
        """
        Return the stored gradient for a given parameter set.

        Parameters
        ----------
        x : np.array
            Parameter set.

        Returns
        -------
        np.array
            Value of the gradient.
        """
        key = str(x)
        gradient = self.gradients.pop(key)
        if np.any(np.isnan(gradient)) or np.any(np.isinf(gradient)):
            # TODO: is simply a warning sufficient?
            gradient[
                np.isnan(gradient)
            ] = 1e-10  # Most probably at boundary of Quantity
            gradient[
                np.isinf(gradient)
            ] = 1e-10  # Most probably at boundary of Quantity
        return gradient

    def fct_to_min(
        self, input_parameters: Union[np.ndarray, tf.constant]
    ) -> Union[np.ndarray, tf.constant]:
        """
        Wrapper for the goal function.

        Parameters
        ----------
        input_parameters : [np.array, tf.constant]
            Vector of parameters in the optimizer friendly way.

        Returns
        -------
        [np.ndarray, tf.constant]
            Value of the goal function. Float if input is np.array else tf.constant
        """
        pars = []
        # We use zip to create pairs of Quantity objects and their new bare numeric parameter values.
        # Then we compute the new physical value and store it in the status.
        for par, y in zip(self.pmap.get_parameters(), np.array(input_parameters)):
            pars.append(par.get_other_value(y).tolist())
        self.optim_status["params"] = pars
        if isinstance(input_parameters, np.ndarray):
            current_params = tf.constant(input_parameters)
            goal = self.goal_run(current_params)
            self.optim_status["goal"] = float(goal)
            goal = float(goal)
        else:
            current_params = input_parameters
            goal = self.goal_run(current_params)
            self.optim_status["goal"] = float(goal)
        self.log_parameters(input_parameters)
        return goal

    def fct_to_min_autograd(self, x):
        """
         Wrapper for the goal function, including evaluation and storage of the
         gradient.

        Parameters
         ----------
         x : np.array
             Vector of parameters in the optimizer friendly way.

         Returns
         -------
         float
             Value of the goal function.
        """
        current_params = tf.constant(x)
        goal, grad = self.goal_run_with_grad(current_params)
        if isinstance(grad, tf.Tensor):
            grad = grad.numpy()
        gradients = grad.flatten()
        for i in tf.where(gradients == 0).numpy().tolist():
            warnings.warn(
                f"{self.pmap.get_key_from_scaled_index(i[0])} has no gradient. This might indicate no usage for current experiment.",
                Warning,
            )
        self.gradients[str(current_params.numpy())] = gradients
        # We use zip to create pairs of Quantity objects and their new bare numeric parameter values.
        # Then we compute the new physical value and store it in the status.
        self.optim_status["params"] = [
            par.get_other_value(y) for par, y in zip(self.pmap.get_parameters(), x)
        ]
        self.optim_status["gradient"] = gradients.tolist()
        last_goal = float(goal)
        self.optim_status["goal"] = last_goal
        self.log_parameters(current_params)
        return goal


class BaseLogger:
    def __init__(self):
        pass

    def start_log(self, opt, logdir):
        self.logdir = logdir

    def log_parameters(self, evaluation, optim_status):
        pass

    def end_log(self, opt, logdir):
        pass


class BestPointLogger(BaseLogger):
    pass


class TensorBoardLogger(BaseLogger):
    def __init__(self):
        super().__init__()
        self.opt_map = []
        self.writer: None
        self.store_better_iterations_only = True
        self.best_iteration = np.inf

    def write_params(self, params, step=0):
        if not len(self.opt_map) == len(params):
            raise Exception(
                f"C3:Error: Different number of elements in opt_map and params. {len(self.opt_map)} vs {len(params)}"
            )
        for i in range(len(self.opt_map)):
            for key in self.opt_map[i]:
                if type(params[i]) is float or np.float64:
                    tf.summary.scalar(key, float(params[i]), step=step)
                elif len(params[i]) == 1:
                    tf.summary.scalar(key, float(params[i][0]), step=step)
                else:
                    for jj in range(len(params[i])):
                        tf.summary.scalar(
                            key + "_" + str(jj), float(params[i][jj]), step=step
                        )

    def set_logdir(self, logdir):
        print("new Tensorboard Logdir")
        self.logdir = logdir

    def start_log(self, opt, logdir):
        self.opt_map = opt.pmap.get_opt_map()
        print("create log at", logdir)
        self.writer = tf.summary.create_file_writer(
            logdir=logdir,
        )

        with self.writer.as_default():
            self.write_params(opt.pmap.get_parameters())
            tf.summary.text(
                "Parameters",
                hjson.dumpsJSON(
                    opt.pmap.asdict(instructions_only=False),
                    indent=2,
                    default=hjson_encode,
                ),
                step=0,
            )

        self.writer.flush()
        hparams = dict()
        for k, hpar in opt.pmap.get_not_opt_params().items():
            val = hpar.numpy()
            if len(val.shape) < 1 or val.shape[0] < 2:
                hparams[k] = float(hpar)
            else:
                for i, v in enumerate(val.tolist()):
                    hparams[f"{k}_{i}"] = v
        with self.writer.as_default():
            hp.hparams(hparams)
        self.writer.flush()

    def log_parameters(self, evaluation, optim_status):
        if self.store_better_iterations_only:
            if optim_status["goal"] > self.best_iteration:
                return
            else:
                self.best_iteration = optim_status["goal"]
        opt_status = copy.deepcopy(optim_status)
        with self.writer.as_default():
            self.write_params(opt_status.pop("params"), evaluation)
            tf.summary.scalar("goal", opt_status.pop("goal"), step=evaluation)

            if "gradient" in opt_status:
                tf.summary.histogram(
                    "gradient",
                    tf.clip_by_value(opt_status.pop("gradient"), -3, 3),
                    step=evaluation,
                )

            opt_status.pop("time")
            for k, v in opt_status.items():
                if type(v) is float:
                    tf.summary.scalar(k, v, step=evaluation)
                elif type(v) is list:
                    tf.summary.histogram(k, v, step=evaluation)
                else:
                    # print(k, v)
                    raise Warning(
                        f"Elements of type {type(v)}, here {k}, are not yet implemented to be logged "
                    )
            self.writer.flush()



"""
testing module quick setup class
"""

import numpy as np
import pytest
import pickle
from c3.experiment import Experiment

exp = Experiment()
exp.load_quick_setup("test/quickstart.hjson")
pmap = exp.pmap
model = pmap.model
generator = pmap.generator


@pytest.mark.integration
def test_exp_quick_setup_freqs() -> None:
    """
    Test the quick setup.
    """
    print(pmap.instructions.keys())
    qubit_freq = model.subsystems["Q1"].params["freq"].get_value()
    gate = pmap.instructions["rx90p[0]"]
    carrier_freq = gate.comps["d1"]["carrier"].params["freq"].get_value()
    offset = gate.comps["d1"]["gaussian"].params["freq_offset"].get_value()
    assert qubit_freq == carrier_freq + offset


@pytest.mark.integration
def test_generator() -> None:
    gen_signal = pmap.generator.generate_signals(pmap.instructions["rx90p[0]"])
    with open("test/quick_data.pickle", "rb") as quickfile:
        test_data = pickle.load(quickfile)
    np.testing.assert_allclose(gen_signal["d1"]["ts"], test_data["d1"]["ts"], rtol=1e-8)
    np.testing.assert_allclose(
        actual=gen_signal["d1"]["values"].numpy(),
        desired=test_data["d1"]["values"].numpy(),
        rtol=1e-8,
    )
